import datetime
import json
import re
import functools
import typing


def _raise_value_error(*args, **kwargs):
    """Simply raise a value error."""
    raise ValueError()


def _unhandled_exceptions_return(return_value: bool) -> typing.Callable:
    """
    Create a decorator that converts all unhandled exceptions to the given
    return value.
    """

    def decorator(func: typing.Callable) -> typing.Callable:
        """Decorate the given input function."""

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            """
            Convert all non keyboard interupt exceptions to the given
            value
            """
            try:
                return func(*args, **kwargs)
            except KeyboardInterrupt as err:
                raise err
            except Exception as err:
                return return_value
        return wrapper

    return decorator


class JsonStringChecker():
    """Checker for validating that a string is json deserializable."""

    name = 'json'

    @_unhandled_exceptions_return(False)
    def __call__(self, value: typing.Any) -> bool:
        """Confirm the given value can be deserialized to json."""
        json.loads(value, parse_constant=_raise_value_error)
        return True


class JsonArrayStringChecker():
    """
    Checker for validating that a string is json deserializable and
    deserializes into an array.
    """

    name = 'json_array'

    @_unhandled_exceptions_return(False)
    def __call__(self, value: typing.Any) -> bool:
        """Confirm the given value can be json deserialized to an array."""
        result = json.loads(value, parse_constant=_raise_value_error)
        return isinstance(result, list)


class JsonObjectStringChecker():
    """
    Checker for validating that a string is json deserializable and
    deserializes into an object.
    """

    name = 'json_object'

    @_unhandled_exceptions_return(False)
    def __call__(self, value: typing.Any) -> bool:
        """Confirm the given value can be json deserialized to an object."""
        result = json.loads(value, parse_constant=_raise_value_error)
        return isinstance(result, dict)


class ChoiceChecker():
    """Checker for validating that a value is in a set of choices."""

    name = 'choice'

    def __init__(self, options: list = None):
        self._options = set(options)

    @_unhandled_exceptions_return(False)
    def __call__(self, value: typing.Any) -> bool:
        """Confirm the given value is a valid choice"""
        return value in self._options


class IsoTimestampChecker():
    """Checker for validating that a value is an iso timestamp."""

    name = 'iso_timestamp'

    @_unhandled_exceptions_return(False)
    def __call__(self, value: typing.Any) -> bool:
        """Confirm the given value is a valid iso timestamp."""
        datetime.datetime.fromisoformat(value)
        return True


class MaxChecker():
    """Checker for validating that a value is less than a given value."""

    name = 'maximum'

    def __init__(
            self, limit: typing.Union[str, int, float], inclusive: bool = True
    ):
        """
        Initialize the checker with its configured values.

        :param limit:
            The maximum value to allow.
        :param inclusive:
            Whether to allow the input value to be be exactly the limit or
            if the limit should be exclusive.
        """
        self._limit = limit
        self._inclusive = inclusive

    @_unhandled_exceptions_return(False)
    def __call__(self, value: typing.Any) -> bool:
        """Confirm that a given value is less than the specified limit."""
        return value <= self._limit if self._inclusive else value < self._limit


class MinChecker():
    """Checker for validating that a value is greater than a given value."""

    name = 'minimum'

    def __init__(
            self, limit: typing.Union[str, int, float], inclusive: bool = True
    ):
        """
        Initialize the checker with its configured values.

        :param limit:
            The minimum value to allow.
        :param inclusive:
            Whether to allow the input value to be be exactly the limit or
            if the limit should be exclusive.
        """
        self._limit = limit
        self._inclusive = inclusive

    @_unhandled_exceptions_return(False)
    def __call__(self, value: typing.Any) -> bool:
        """Confirm that a given value is greater than the specified limit."""
        return value >= self._limit if self._inclusive else value > self._limit


class RegexChecker():
    """Checker for validating that a value meets the given regex."""

    name = 'regex'

    def __init__(self, regex: str = None, ignore_case=False):
        flags = re.IGNORECASE if ignore_case else 0
        self._regex = re.compile(regex, flags)

    @_unhandled_exceptions_return(False)
    def __call__(self, value: typing.Any) -> bool:
        """Confirm the given value matches the regex."""
        return bool(self._regex.match(value))


class EmailChecker(RegexChecker):
    """Checker for validating that a value is a valid email."""

    name = 'email'

    def __init__(self):
        super().__init__(
            regex=r'^[A-Z0-9._%+-]+@[A-Z0-9.-]+\.[A-Z]{2,63}$',
            ignore_case=True
        )


class UuidChecker(RegexChecker):
    """Checker for validating that a value is a valid email."""

    name = 'uuid'

    def __init__(self):
        pat = r'^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$'
        super().__init__(regex=pat, ignore_case=True)


class MaxLengthChecker():
    """Checker for validating the max length of a string."""

    name = 'maximum_length'

    def __init__(self, limit: int):
        """
        Initialize the checker with its configured values.

        :param limit:
            The maximum value length to allow.
        """
        self._limit = limit

    @_unhandled_exceptions_return(False)
    def __call__(self, value: typing.Any) -> bool:
        """Confirm the given value is within the max length."""
        return len(value) <= self._limit


CHECKERS = [
    ChoiceChecker,
    EmailChecker,
    IsoTimestampChecker,
    JsonArrayStringChecker,
    JsonObjectStringChecker,
    JsonStringChecker,
    MaxChecker,
    MaxLengthChecker,
    MinChecker,
    RegexChecker,
    UuidChecker,
]
CHECKER_MAP = {c.name: c for c in CHECKERS}



import typing
import dataclasses

@dataclasses.dataclass(frozen=True)
class Field:
    """
    The definition of a field within a schema.

    :vartype name: str
    :ivar name:
        The name of the field.
    :vartype type_: str
    :ivar type_:
        The type of the field. Allowed values include: "boolean", "integer",
        "float", and "string".
    :vartype description: str
    :ivar description:
        A detailed description of the field.
    :vartype required: bool
    :ivar required:
        Whether the field is required to be present.
    :vartype required: bool
    :ivar nullable:
        Whether the field should accept null values.
    :vartype checks: list
    :ivar checks:
        A list of tuples, each tuple containing the name of a the check
        and a callable that can be used to evaluate the check.
    :vartype labels: dict
    :ivar labels:
        A space for unstructured information regarding the field.
    """

    name: str
    type_: str
    description: str
    required: bool
    nullable: bool
    checks: list
    labels: dict


@dataclasses.dataclass(frozen=True)
class Check:
    """
    The definition of a check within a field.

    :vartype name: str
    :ivar name:
        The name of the field.
    :vartype handler: typing.Callable
    :ivar handler:
        A callable object that processes a single value, returning True
        if the value passes the check or False if it fails.
    :vartype args: dict
    :ivar args:
        The arguments used to configure the handler.
    """

    name: str
    handler: typing.Callable
    args: dict



import dataclasses
import functools
import inspect
import typing
import yaml

from horkos import _casters
from horkos import _checkeromatic
from horkos import _definitions
from horkos import errors

Schemaable = typing.Union[str, dict, 'Schema', typing.TextIO]


@dataclasses.dataclass(frozen=True)
class Schema:
    """
    The schema of a data set.

    :vartype name: str
    :ivar name:
        The name of the data set the schema describes.
    :vartype description: str
    :ivar description:
        A detailed description of the data set.
    :vartype labels: dict
    :ivar labels:
        An unstructured area of the schema for storing organization
        specific information relating to the data set.
    :vartype fields: typing.Dict[str, horkos.Field]
    :ivar fields:
        A dictionary mapping field names to field objects
    """

    name: str
    description: str
    labels: dict
    fields: typing.Dict[str, 'horkos.Field']

    def cast(self, record: dict) -> dict:
        """
        Cast all the values contained within the record to the types declared
        within the schema. All undeclared fields will be removed from the
        record.

        :param record:
            The record to cast using the schema. This is expected to be
            a dictionary mapping field names to field values.
        :return:
            The record with values cast to the correct types.
        """
        return _casters.cast(record, self.fields)

    def find_errors(self, record: dict) -> list:
        """
        Run all field checks against the record, identifying all failed
        checks. Returns a list of errors if any of the checks have failed
        or an empty list if all passed.

        :param record:
            The record to inspect for errors.
        :return:
            A list of errors associated with the record.
        """
        error_list = []
        for field in self.fields.values():
            if field.required and field.name not in record:
                error_list.append(f'{field.name} is a required field')
                continue
            value = record.get(field.name)
            if value is None:
                if not field.nullable:
                    error_list.append(f'{field.name} cannot be null')
                continue
            for check in field.checks:
                if not check.handler(value):
                    error_list.append(
                        f'value of "{value}" in {field.name} did not '
                        f'pass {check.name} check'
                    )
        return error_list


    def process(self, record: dict) -> dict:
        """
        Process a record against the schema, casting each field of the
        record to its specified type and validating the the given value
        passes all of the checks declared in the schema. If the fields
        cannot be successfully cast or any of the checks failed a
        RecordValidationError exception will be raised.

        :param record:
            The record to process against the schema. This should be
            a dictionary mapping field names to field values.
        :return:
            The processed record with values cast and validated.
        """
        casted = self.cast(record)
        error_list = self.find_errors(casted)
        if error_list:
            raise errors.RecordValidationError(
                f'Check errors - {", ".join(error_list)}'
            )
        return casted


def _schema_from_dict(
        raw_schema: dict,
        strict: bool = True,
        custom_checkers: typing.Dict[str, typing.Callable] = None
) -> Schema:
    """
    Create a schema from a raw schema dictionary.

    :param raw_schema:
        A raw schema matching the expected json/yaml schema format.
    :param strict:
        Whether to operate in strict mode. If True descriptions will be
        required.
    :param custom_checkers:
        A dictionary of custom checkers. This dictionary should map the name
        of the check to a function that generates the check function
    :return:
        The resulting schema.
    """
    standardized = _standardize_schema(raw_schema, strict=strict)
    checkers = _checkeromatic.CHECKER_MAP.copy()
    checkers.update(custom_checkers or {})
    fields = {}
    raw_fields = standardized['fields']
    for field_name, field_body in raw_fields.items():
        checks = []
        for i, raw_check in enumerate(field_body['checks']):
            position = f'fields.{field_name}.checks[{i}]'
            check_name = raw_check['name']
            if check_name not in checkers:
                raise errors.SchemaValidationError(
                    f'In "{position}" check name "{check_name}" is unknown.'
                )
            check = _create_check_func(
                checkers[check_name],
                raw_check['args'],
                position
            )
            checks.append(
                _definitions.Check(check_name, check, raw_check['args'])
            )
        fields[field_name] = _definitions.Field(
            field_name,
            field_body['type'],
            field_body['description'],
            field_body['required'],
            field_body['nullable'],
            checks,
            field_body['labels'],
        )
    return Schema(
        standardized['name'],
        standardized['description'],
        standardized['labels'],
        fields,
    )

def _load_yaml(
        filepath_or_handle: typing.Union[str, typing.TextIO]
) -> dict:
    """
    Read the given yaml file.

    :param filepath_or_handle:
        The path to the yaml file to convert into a schema.
    :return:
        The resulting schema.
    """
    if not isinstance(filepath_or_handle, str):
        return yaml.safe_load(filepath_or_handle)
    with open(filepath_or_handle) as f:
        return yaml.safe_load(f)


def _create_check_func(
        check: typing.Callable, args: dict, position: str
) -> typing.Callable:
    """
    Create the actual check function, baking the args into it. The returned
    function should only expect a single positional argument to check.

    :param check:
        The base check. This is either a class that will be initialized
        with the given arguments, or a function that will be partially filled
        with the given args.
    :param args:
        The arguments to initialize the check with.
    :param position:
        The position of the check within the schema. This is used to provide
        more clear error messages.
    :return:
        A callable that expects a single positional argument that it will
        check.
    """
    try:
        if inspect.isclass(check):
            return check(**args)
        signature = inspect.signature(check)
        signature.bind('value', **args)
    except TypeError:
        raise errors.SchemaValidationError(
            f'Invalid arguments in "{position}"'
        )
    return functools.partial(check, **args)


def _standardize_schema(raw_schema: dict, strict: bool = True) -> dict:
    """
    Standardize the given raw_schema with all optional fields filled out
    with defaults. If the raw schema is not valid a SchemaValidationError will
    be raised.

    :param raw_schema:
        The schema to standardize.
    :param strict:
        Whether to operate in strict mode. If True descriptions will be
        required.
    :return:
        The standardized schema.
    """
    standardized = {}
    top = [
        ('name', str, True),
        ('description', str, strict),
        ('labels', dict, False),
        ('fields', dict, True),
    ]
    for key, type_, required in top:
        if key not in raw_schema and required:
            raise errors.SchemaValidationError(f'"{key}" is required')
        value = raw_schema[key] if key in raw_schema else type_()
        if not isinstance(value, type_):
            raise errors.SchemaValidationError(
                f'"{key}" must be of type {type_.__name__}'
            )
        standardized[key] = value
    standardized['fields'] = {
        field_id: _standardize_field(raw_field, field_id, strict=strict)
        for field_id, raw_field in raw_schema['fields'].items()
    }
    return standardized


def _standardize_field(
        raw_field: dict, field_id: str, strict: bool = True
) -> dict:
    """
    Standardize a given raw field with all the optional parts filled out
    with defaults. If the raw field can't be validated a SchemaValidationError
    will be raised.

    :param raw_field:
        The raw field to standardize.
    :param field_id:
        The id of the field being standardized.
    :param strict:
        Whether to operate in strict mode. If True descriptions will be
        required.
    :return:
        The standardized field.
    """
    standardized = {}
    top = [
        ('type', str, True, None),
        ('description', str, strict, str),
        ('labels', dict, False, dict),
        ('checks', list, False, list),
        ('required', bool, False, lambda: True),
        ('nullable', bool, False, bool),
    ]
    for key, type_, required, default in top:
        working = f'fields.{field_id}.{key}'
        if key not in raw_field and required:
            raise errors.SchemaValidationError(f'"{working}" is required')
        value = raw_field[key] if key in raw_field else default()
        if not isinstance(value, type_):
            raise errors.SchemaValidationError(
                f'"{working}" must be of type {type_.__name__}'
            )
        standardized[key] = value
    standardized['checks'] = [
        _standardize_check(raw_check, f'fields.{field_id}.check[{i}]')
        for i, raw_check in enumerate(raw_field.get('checks') or [])
    ]
    return standardized


def _standardize_check(
        raw_check: typing.Union[str, dict], prefix: str
) -> dict:
    """
    Convert a check into the standard format. If the check cannot be
    standardize a schema validation error is raised.

    :param raw_check:
        The raw_check to standardize.
    :param prefix:
        The document position prefix to integrate into error messages.
    :return:
        A standard check dictionary.
    """
    if isinstance(raw_check, str):
        return {'name': raw_check, 'args': {}}
    if isinstance(raw_check, dict):
        if 'name' not in raw_check:
            raise errors.SchemaValidationError(
                f'"{prefix}.name" is required'
            )
        if not isinstance(raw_check['name'], str):
            raise errors.SchemaValidationError(
                f'"{prefix}.name" must be a string'
            )
        if 'args' in raw_check and not isinstance(raw_check['args'], dict):
            raise errors.SchemaValidationError(
                f'"{prefix}.args" must be a dictionary'
            )
        return {'name': raw_check['name'], 'args': raw_check.get('args', {})}
    raise errors.SchemaValidationError(
        f'"{prefix}" must be a string or dictionary'
    )


def load_schema(
        schema: Schemaable,
        strict: bool = False,
        custom_checkers: typing.Dict[str, typing.Callable] = None
) -> Schema:
    """
    Load a schema from a file or existing schema object.

    :param schema:
        The schema to load. If a string is given, it is assumed to be a path
        to a `.json` or `.yaml` file defining the schema, alternatively a file
        handle can be passed directly. If a dictionary is given it will be
        assumed to be equivalent to the contents of a schema file.
    :param strict:
        Whether to operate in strict mode. If True descriptions will be
        required.
    :param custom_checkers:
        A dictionary of custom checkers. This dictionary should map the name
        of the check to a function that generates the check function. This
        can be used to extend checks beyond those defined within `horkos`.
    :return:
        A Schema object.
    """
    working = schema
    custom_checkers = custom_checkers or {}
    if isinstance(working, Schema):
        return working
    if not isinstance(working, dict):
        working = _load_yaml(working)
    return _schema_from_dict(
        working, strict=strict, custom_checkers=custom_checkers
    )



from horkos._cataloger import Catalog
from horkos._definitions import Check
from horkos._definitions import Field
from horkos._schemaomatic import load_schema
from horkos._schemaomatic import Schema


__all__ = ('Catalog', 'Check', 'Field', 'load_schema', 'Schema')



import io
import typing

from horkos import _schemaomatic


class Catalog:
    """
    A collection of schemas.

    :vartype schemas: typing.Dict[str, horkos.Schema]
    :ivar schemas:
        All of the schemas in the catalog.
    """

    def __init__(
            self,
            schemas: typing.List[_schemaomatic.Schemaable] = None,
            strict: bool = True,
    ):
        """
        Initialize the catalog of data schemas.

        :param schema_files:
            Yaml file paths of schema files to include in the catalog.
        :param schemas:
            Existing schema objects to include in the catalog.
        :param strict:
            Whether to operate in strict mode. If True descriptions will be
            required.
        """
        self.strict = strict
        schemas = schemas or []
        all_schemas = [
            _schemaomatic.load_schema(s, strict=strict) for s in schemas
        ]
        self.schemas = {s.name: s for s in all_schemas}

    def cast(self, name: str, record: dict) -> dict:
        """
        Cast the fields of the record to the types described in its schema.
        All undeclared fields will be removed from the record.

        :param name:
            The name of the schema associated with the record.
        :param record:
            The record to cast using the schema. This is expected to be
            a dictionary mapping field names to field values.
        :return:
            The record with values cast to the correct types.
        """
        schema = self._pull_schema(name)
        return schema.cast(record)

    def find_errors(self, name: str, record: dict) -> list:
        """
        Identify any failed checks for the given record.

        Run all field checks against the record, identifying all failed
        checks. Returns a list of errors if any of the checks have failed
        or an empty list if all passed.

        :param name:
            The name of the schema associated with the record.
        :param record:
            The record to inspect for errors.
        :return:
            A list of errors associated with the record.
        """
        schema = self._pull_schema(name)
        return schema.find_errors(record)

    def process(
            self,
            name: str,
            record: dict,
    ) -> dict:
        """
        Process a record against a named schema from the catalog.

        Each field within the record will be cast to the type specified in
        the schema and the resulting value validated against the checks
        defined within the schema. If any of the fields
        cannot be successfully cast or any of the checks fail a
        RecordValidationError exception will be raised.

        :param name:
            The name of the schema to process the record against.
        :param record:
            The record to process against the schema. This should be
            a dictionary mapping field names to field values.
        :return:
            The processed record with values cast and validated.
        """
        schema = self._pull_schema(name)
        return schema.process(record)

    def _pull_schema(self, name: str) -> _schemaomatic.Schema:
        """
        Pull the schema with the given name, if no matching schema exists
        raise a value error indicating as much.
        """
        schema = self.schemas.get(name)
        if schema is None:
            raise ValueError(f'No schema exists for "{name}"')
        return schema

    def update(self, schema: _schemaomatic.Schemaable):
        """
        Update the catalog with the given schema.

        :param schema:
            The schema to add/update to the catalog.
        """
        to_add = _schemaomatic.load_schema(schema, strict=self.strict)
        self.schemas[to_add.name] = to_add



"""
Critique a schema or catalog of schemas. This identifies common mistakes in
field design and potential issues between schemas within a catalog.
"""
import argparse
import os
from os import path
import re
import sys
import typing

import yaml

from horkos import errors
from horkos import _cataloger
from horkos import critiquer
from horkos import _schemaomatic

SCHEMA_FILE_RE = re.compile(r'.*\.(yaml|yml|json)$')


def configure_parser(parser: argparse.ArgumentParser):
    """Configure the given parser to work for the critique command."""
    parser.add_argument(
        '--schema', '-s',
        dest='schema_file',
        help='The schema file to critique.'
    )
    parser.add_argument(
        '--catalog', '-c',
        action='append',
        dest='catalog_files',
        help=(
            'A path defining the catalog of schemas. If a directory is given '
            'all yaml files contained within will be used as schemas in the '
            'catalog. If used without the `--schema` flag the catalog itself '
            'will be critiqued, otherwise the flagged schema will be '
            'critiqued within the context of the given catalog.'
        )
    )


def _expand_directories(file_paths: typing.List[str]) -> typing.List[str]:
    """
    Expand any directories in the given set of paths replacing them with all
    of the yaml files they contain.
    """
    result = []
    for file_path in file_paths:
        if not path.exists(file_path):
            print(f'ERROR: {file_path} does not exist.', file=sys.stderr)
            sys.exit(1)
        if not path.isdir(file_path):
            if not SCHEMA_FILE_RE.match(file_path):
                print(
                    f'ERROR: {file_path} is not a yaml or json file.',
                    file=sys.stderr
                )
                sys.exit(1)
            result.append(file_path)
            continue
        for root, _, files in os.walk(file_path):
            result.extend(
                path.abspath(path.join(root, f))
                for f in files
                if SCHEMA_FILE_RE.match(f)
            )
    return result


def _load_schemas(schema_file: str) -> typing.List[_schemaomatic.Schema]:
    """Load the given schema file returnings all schemas it contains."""
    if not path.isfile(schema_file):
        print(f'ERROR: {schema_file} does not exist.', file=sys.stderr)
        sys.exit(1)
    with open(schema_file) as f:
        contents = f.read()
    return [
        _schemaomatic.load_schema(doc) for doc in yaml.safe_load_all(contents)
    ]


def _assemble_catalog(catalog_files: typing.List[str], omit: str = None):
    """Assemble a catalog using the given catalog files."""
    if omit in catalog_files:
        print(
            f'ERROR: {omit} is given for both --schema and --catalog',
            file=sys.stderr,
        )
        sys.exit(1)
    expanded_catalog = set(_expand_directories(catalog_files))
    if omit:
        expanded_catalog -= {path.abspath(omit)}
    catalog = _cataloger.Catalog()
    for file_path in expanded_catalog:
        with open(file_path) as f:
            contents = f.read()
        for doc in yaml.safe_load_all(contents):
            catalog.update(doc)
    return catalog


def _schema_critique(schema_file: str) -> typing.List[critiquer.Critique]:
    """Critique a schema."""
    critiques = []
    for schema in _load_schemas(schema_file):
        critiques.extend(critiquer.critique_schema(schema))
    return critiques


def _relative_critique(
        schema_file: str, catalog_files: typing.List[str]
) -> typing.List[critiquer.Critique]:
    """Critique the schema within the context of the catalog."""
    schemas = _load_schemas(schema_file)
    catalog = _assemble_catalog(catalog_files, omit=schema_file)
    for schema in schemas:
        if schema.name not in catalog.schemas:
            continue
        print(
            f'A schema named {schema.name} is already in the catalog',
            file=sys.stderr,
        )
        sys.exit(1)
    critiques = []
    for schema in _load_schemas(schema_file):
        critiques.extend(critiquer.critique_addition(schema, catalog))
    return critiques


def _catalog_critique(
        catalog_files: typing.List[str]
) -> typing.List[critiquer.Critique]:
    """Critique the catalog."""
    catalog = _assemble_catalog(catalog_files)
    return critiquer.critique_catalog(catalog)


def main(
        schema_file: str,
        catalog_files: typing.List[str],
        **kwargs,
):
    """Critique the given schema / catalog."""
    if not schema_file and not catalog_files:
        print('--schema or --catalog must be specified', file=sys.stderr)
        sys.exit(1)
    if schema_file and not catalog_files:
        critiques = _schema_critique(schema_file)
    if schema_file and catalog_files:
        critiques = _relative_critique(schema_file, catalog_files)
    if not schema_file and catalog_files:
        critiques = _catalog_critique(catalog_files)
    for critique in critiques:
        print(critique)
    if critiques:
        sys.exit(1)
